angular.module("beamng.apps")
/**
 * The base service for the future variant apps
 */
.factory("ModMinBase", ["TachoBase", function(TachoBase) {
  let outerThis = this;

  /**
   *
   */
  class GaugeBuilder extends TachoBase.GaugeBuilder {
    /**
     * Constructor
     * @param {number} baseRadius - the radius of the gauge at its maximum point
     * @param {ColourScheme} colourScheme - the colour scheme to apply to the gauge
     * @param {Range} angleRange - the range of angles of the dial's arc overall
     */
    constructor(baseRadius, colourScheme, angleRange) {
      super();
      // @type number
      this.baseRadius = baseRadius;
      // @type ColourScheme
      this.colourScheme = colourScheme;
      // @type Range
      this.angleRange = angleRange;
    }

    /**
     * @Override
     * @param {Number} dialRadius - the radius of the middle dial
     * @param {Vector} centre - the centre point of the dial
     * @param {Number} majorLength - the length of the pointer on the major side
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createPointer(dialRadius, centre, majorLength) {
      let accent = this.colourScheme.ACCENT;
      let primaryStyle = TachoBase.Style.unitStrokeFill(accent.secondary);

      this.pointer = new TachoBase.SimplePointer(
          new TachoBase.Rectangle(6, majorLength, primaryStyle),
          new TachoBase.Rectangle(6, majorLength/3, primaryStyle),
          new TachoBase.Circle(dialRadius, new TachoBase.Style(accent.primary, 0, true), centre)
      );
      return this;
    }

    /**
     * Creates the GaugeTicks instance for the gauge
     * @param {number} numTicks - the number of ticks to generate
     * @param {number} tickSpacing - the spacing between two ticks, in terms of value
     * @param {number} arcWidth - the width of the arc
     * @param {number} minTickVal - the lowest tick's value. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createTicks(numTicks, tickSpacing, arcWidth, minTickVal) {
      this.gaugeTicks = new TachoBase.GaugeTicks(
          numTicks,
          tickSpacing,
          3,
          2,
          this.colourScheme.ACCENT,
          // +-2 to not go over bounds
          new TachoBase.Range(this.baseRadius - arcWidth + 2, this.baseRadius - 2),
          this.angleRange,
          true,
          minTickVal
      );
      return this;
    }

    /**
     * Creates the GaugeNumbers instance for the gauge
     * @param {number} numNumbers - the number of numbers that will fill the major points on the
     * tick gauge
     * @param {number} fontSize - the size of the font for the numbers
     * @param {number} numbersRadius - the radius of the arc that the numbers will be spaced around
     * @param {number} spacing - the spacing between two numbers
     * @param {number} [startNum] - the starting number. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createNumbers(numNumbers, fontSize, numbersRadius, spacing, startNum) {
      this.gaugeNumbers = new TachoBase.GaugeNumbers(
          numNumbers,
          spacing,
          this.angleRange,
          new TachoBase.Font(outerThis.FONT_NAME, fontSize, true),
          numbersRadius,
          this.colourScheme.TEXT.secondary,
          startNum
      );
      return this;
    }
  }


  const MT_WIDTH = {
    // The current RPM arc that goes around the outside
    RPM_ARC: 30,
    // The interior's stroke between itself and the RPM background
    INNER_STROKE: 5,
    SPEED_BOX_WIDTH: 100,
    SPEED_BOX_HEIGHT: 60
  };

  const MT_RADII = {
    // The radius of the whole UI
    BASE: 150,
    // The radius of the pointer's dial
    RPM_DIAL_RADIUS: 20
  };
  // The radius of the RPM arc that goes around the outside of the gauge
  MT_RADII.RPM_ARC = MT_RADII.BASE - MT_WIDTH.RPM_ARC/2;
  // The radius of the inner (stroked) part of the RPM arc
  MT_RADII.RPM_INNER = MT_RADII.BASE - MT_WIDTH.RPM_ARC;
  // The radius the RPM numbers are displayed around
  MT_RADII.RPM_NUMBERS = MT_RADII.RPM_ARC - MT_WIDTH.RPM_ARC;


  /**
   * The concrete class representation of this app
   */
  class VelocityTachoApp extends TachoBase.TachometerApp {
    // The width of stroke effects for various components
    static WIDTH = MT_WIDTH;

    /*
     * The centre radius values for various components.
     * It should be noted that the radii values are with respect to an arc that they represent.
     * So, a radius value is not the radius of that element, but the radius of the arc around the
     * centre of the UI it will be drawn as.
     * Any stroke effect adds thickness to a radius
     */
    static RADII = MT_RADII;

    // The RPM spacing between two ticks
    static TICK_SPACING = 500;


    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityTachoApp.RADII.BASE, document, bngApi, UiUnits);
      this.needsRedraw = false;
    }

    /**
     * @Override
     * @inheritDoc
     */
    updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel) {
      let prevState = this.collateBGVars();
      super.updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel);
      let newState = this.collateBGVars();

      // If any variables for the background canvas have changed, the background canvas
      // requires a re-draw
      if (!prevState.every(function(v, i) { return v === newState[i]}))
        this.needsRedraw = true;
    }

    /**
     * @returns {[*]} the list of all variables that drawn on the background canvas. This
     * includes the gear index, the lights' state and the parking brake
     */
    collateBGVars() {
      return [this.currGearIndex, this.lightsState, this.pBrakeEngaged];
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      // Create concrete class of base class builder
      class TachoVelocityGaugeBuilder extends outerThis.GaugeBuilder {
        constructor(baseRadius, colourScheme, angleRange) {
          super(baseRadius, colourScheme, angleRange);
        }

        /**
         * Creates the GaugeArcs instance for the gauge
         * @param arcRadius
         * @param arcWidth
         * @param redlineAngle
         * @returns {TachoVelocityGaugeBuilder} the GaugeBuilder instance
         */
        createArcs(arcRadius, arcWidth, redlineAngle) {
          this.gaugeArcs = new TachoBase.GaugeArcs(
              null,
              new TachoBase.Arc(
                  arcRadius,
                  new TachoBase.Range(redlineAngle, this.angleRange.upper),
                  new TachoBase.Style("#ff2020", arcWidth)
              ),
              null
          );
          return this;
        }
      }

      this.gauge = new TachoVelocityGaugeBuilder(this.radius, this.colourScheme, this.RPM_ANGLE_RANGE)
          .createPointer(
              VelocityTachoApp.RADII.RPM_DIAL_RADIUS,
              this.CENTRE,
              // Poke halfway into the RPM arc
              this.radius - VelocityTachoApp.WIDTH.RPM_ARC/2
          )
          // +1 to include 0RPM
          .createTicks(
              Math.floor(this.redLineRPM / VelocityTachoApp.TICK_SPACING) + 1,
              VelocityTachoApp.TICK_SPACING,
              VelocityTachoApp.WIDTH.RPM_ARC
          )
          .createArcs(
              VelocityTachoApp.RADII.RPM_ARC,
              VelocityTachoApp.WIDTH.RPM_ARC,
              this.getRedlineRPMAngle()
          )
          // +1 to include 0
          .createNumbers(this.redLineRPM/1000 + 1, 26, VelocityTachoApp.RADII.RPM_NUMBERS, 1)
          .setCentre(this.CENTRE)
          .build();
    }

    /**
     * @Override
     * @inheritDoc
     * This includes both backgrounds for the speedometer as well as the divider between the
     * RPM gauge and inner part. This also draws the background parts of the RPM (see
     * setupRPM()) and draws the gear in the centre (see drawGear())
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      let ctx = this.BG_CANVAS.get2DContext();

      // The background circle of the whole gauge
      new TachoBase.Circle(
          VelocityTachoApp.RADII.BASE, TachoBase.Style.noStrokeFill("rgb(20, 20, 20)"), this.CENTRE
      ).draw(ctx)

      this.setupRPM();
      this.drawGear();
      this.drawSpeedBackground();
    }

    /**
     * Setups and draws the RPM gauge's background canvas objects.
     * This also draws the inner parts of the RPM pointer as well as generating the RPM ticks
     */
    setupRPM() {
      let ctx = this.BG_CANVAS.get2DContext();
      // Draw the pointer dial as this only needs to be drawn once
      this.gauge.pointer.pointerDial.draw(ctx);
    }

    /**
     * Draws the gear number in the middle of the tachometer
     */
    drawGear() {
      let gearName = TachoBase.TachometerApp.gearIndexToName(this.currGearIndex);
      // Base font size is 40; if there is more >1 digit on the gear then decrease to fit
      let fontSize = 40 / (1 + (gearName.length - 1) * 0.5);

      let ctx = this.CANVAS.get2DContext();

      // this.drawPrimaryText(ctx, gearName, fontSize, this.CENTRE);
      TachoBase.BaseApp.resetFilter(ctx);
    }

    /**
     * Draws the background polygon for the speedometer
     */
    drawSpeedBackground() {
      let ctx = this.BG_CANVAS.get2DContext();

      let tlPoint = this.getSpeedBoxTLPos();
      new TachoBase.Rectangle(MT_WIDTH.SPEED_BOX_WIDTH, MT_WIDTH.SPEED_BOX_HEIGHT, TachoBase.Style.noStrokeFill(this.colourScheme.BASE), tlPoint)
          .draw(ctx);
    }

    /**
     * @returns {Vector} the top left corner point of the speed box rectangle
     */
    getSpeedBoxTLPos() {
      let tlPoint = TachoBase.Vector.copy(this.CENTRE);
      tlPoint.add(new TachoBase.Vector(-MT_WIDTH.SPEED_BOX_WIDTH/2, MT_WIDTH.SPEED_BOX_HEIGHT/2));
      // Move down a little bit
      tlPoint.sub(new TachoBase.Vector(0, -35));
      return tlPoint;
    }

    /**
     * @Override
     * @inheritDoc
     * Draws the RPM gauge and the interior
     */
    draw() {
      super.draw();
      this.drawRPM();
      this.drawInterior();
    }

    /**
     * Draws the interior parts of the tachometer. This includes the speedometer
     */
    drawInterior() {
      // Re-draw background canvas if a redraw is required
      if (this.needsRedraw) {
        this.drawBackgroundCanvas();
        this.needsRedraw = false;
      }

      this.drawGear();

      let convertedSpeed = this.UiUnits.speed(this.speed, this.unit);
      this.drawSpeed(Math.floor(convertedSpeed.val), convertedSpeed.unit);
    }

    /**
     * Draws the speedometer and its units
     * @param {number} speed - the speed of the vehicle
     * @param {string} unit - the unit {@param speed} is in
     */
    drawSpeed(speed, unit) {
      // The spacing between pieces of text and the space between the top of the speed-box and
      // the first piece of text
      const SPACING = 5;
      const SPEED_FONT_SIZE = 40;
      const UNIT_FONT_SIZE = 20;
      let centrePos = this.getSpeedBoxTLPos();
      centrePos.addX(MT_WIDTH.SPEED_BOX_WIDTH/2);

      let ctx = this.CANVAS.get2DContext();

      // Centre the text
      centrePos.addY(SPACING * 5);
      this.drawPrimaryText(ctx, speed.toString(), SPEED_FONT_SIZE, centrePos);
      // Account for the size of the speed text to ensure no-overlap
      centrePos.addY(SPEED_FONT_SIZE / 2 + SPACING);
      this.drawPrimaryText(ctx, unit, UNIT_FONT_SIZE, centrePos);
    }

    /**
     * Draws the RPM gauge. This includes the RPM ticks, pointer and current RPM arc
     */
    drawRPM() {
      let currRPMAngle = this.RPM_ANGLE_RANGE.getLinearRatio(this.currRPM / this.redLineRPM);
      this.gauge.setAngle(currRPMAngle);
      this.gauge.draw(this.CANVAS.get2DContext(), this.currRPM);
    }

    /**
     * Draws centred text in the colour scheme's primary text colour at the position specified
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {string} text - the text to write
     * @param {number} fontSize - the font size in px
     * @param {Vector} position - the position to set the centre of the text to
     */
    drawPrimaryText(ctx, text, fontSize, position) {
      new TachoBase.Text(
          new TachoBase.Font(this.FONT_NAME, fontSize, true),
          text,
          position,
          this.colourScheme.TEXT.primary
      ).draw(ctx);
    }
  }


  const VFI_WIDTH = {
    // The current boost arc that goes around the outside
    BOOST_ARC: 20,
    // The width of each tick
    TICK_WIDTH: 2
  }

  const VFI_RADII = {
    BASE: 75,
    DIAL_RADIUS: 10
  };
  VFI_RADII.BOOST_ARC = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC/2;
  VFI_RADII.BOOST_NUMBERS = (VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC)/2;

  VFI_WIDTH.BOOST_NUMBERS = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC;

  /**
   * The concrete class representation of this app
   */
  class VelocityBoostApp extends TachoBase.BoostApp {
    // The width of stroke effects for various components
    static WIDTH = VFI_WIDTH;

    // The centre radius values for various components
    // Any stroke effect adds thickness to a radius
    static RADII = VFI_RADII;

    // The number of ticks between two numbers on the gauge
    static NUM_TICKS_PER_GAP = 3;
    // The number of numbers on the gauge to display
    static NUM_NUMBERS = 5;

    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityBoostApp.RADII.BASE, document, bngApi, UiUnits);
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      // Create concrete class of base class builder
      class TachoVelocityGaugeBuilder extends outerThis.GaugeBuilder {
        constructor(baseRadius, colourScheme, angleRange) {
          super(baseRadius, colourScheme, angleRange);
        }

        /**
         * Creates the GaugeArcs instance for the gauge
         * @param arcRadius
         * @param arcWidth
         * @param redlineAngle
         * @returns {TachoVelocityGaugeBuilder} the GaugeBuilder instance
         */
        createArcs(arcRadius, arcWidth, redlineAngle) {
          this.gaugeArcs = new TachoBase.GaugeArcs(
              new TachoBase.Arc(
                  arcRadius,
                  // Will be overridden later, so default to no arc
                  new TachoBase.Range(this.angleRange.lower, this.angleRange.lower),
                  new TachoBase.Style(this.colourScheme.ACCENT.secondary, arcWidth)
              )
          );
          return this;
        }
      }


      let unitMaxBoost = this.UiUnits.pressure(this.maxBoost, this.unit).val;
      let floorMaxBoost = Math.floor(unitMaxBoost + 1);
      // Increment between numbers on the gauge
      let numberInc = floorMaxBoost / Math.floor(VelocityBoostApp.NUM_NUMBERS / 2);

      let numTicks = VelocityBoostApp.NUM_TICKS_PER_GAP * (VelocityBoostApp.NUM_NUMBERS-1)
                     + VelocityBoostApp.NUM_NUMBERS;
      let tickSpacing = numberInc / (VelocityBoostApp.NUM_TICKS_PER_GAP + 1);

      this.gauge = new TachoVelocityGaugeBuilder(
          VelocityBoostApp.RADII.BASE,
          this.colourScheme,
          this.BOOST_ANGLE_RANGE
      )
          .createPointer(
              VelocityBoostApp.RADII.DIAL_RADIUS,
              this.CENTRE,
              this.radius - VelocityBoostApp.WIDTH.BOOST_ARC
          )
          // +1 to include 0 boost
          .createTicks(numTicks, tickSpacing, VelocityBoostApp.WIDTH.BOOST_ARC, -unitMaxBoost)
          .createArcs(VelocityBoostApp.RADII.BOOST_ARC, VelocityBoostApp.WIDTH.BOOST_ARC)
          .createNumbers(
              VelocityBoostApp.NUM_NUMBERS,
              10,
              VelocityBoostApp.RADII.BOOST_NUMBERS,
              numberInc,
              -floorMaxBoost
          )
          .setCentre(this.CENTRE)
          .build();
    }

    /**
     * Draws/updates the background canvas
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();

      let ctx = this.BG_CANVAS.get2DContext();

      // The wedge not covered by the gauge area
      new TachoBase.Wedge(VelocityBoostApp.RADII.BASE, this.BOOST_ANGLE_RANGE.invert(), TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE)
          .draw(ctx);

      // The background arc to the RPM numbers
      new TachoBase.Arc(VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.BASE, VelocityBoostApp.WIDTH.BOOST_NUMBERS+1), this.CENTRE)
          .draw(ctx);
      // Draw background arc to boost gauge
      // Note: -1 to reduce overlap when the other canvas draws over it
      new TachoBase.Arc(VelocityBoostApp.RADII.BOOST_ARC, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, VelocityBoostApp.WIDTH.BOOST_ARC), this.CENTRE)
          .draw(ctx);

      // Draw the pointer dial as this only needs to be drawn once
      this.gauge.pointer.pointerDial.draw(ctx);

      this.drawUnit();
    }

    /**
     * Draws the boost's unit
     */
    drawUnit() {
      this.drawUnitText();
    }

    /**
     * Draws the unit's name in text form
     */
    drawUnitText() {
      const FONT_SIZE = 20;
      const SPACING = 10;

      let ctx = this.BG_CANVAS.get2DContext();

      let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE.lower * 0.275);
      centrePos.addY(SPACING);

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, FONT_SIZE, true), this.unit, centrePos, this.colourScheme.TEXT.primary)
          .draw(ctx);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw() {
      super.draw();
      let ctx = this.CANVAS.get2DContext();

      let boost = this.UiUnits.pressure(this.currBoost, this.unit).val;

      // Normalise from range [-1,1] to [0,1]
      let normBoost = (this.currBoost/this.maxBoost + 1)/2;
      let currBoostAngle = this.BOOST_ANGLE_RANGE.getLinearRatio(normBoost);

      this.gauge.setAngle(currBoostAngle);
      this.gauge.draw(ctx, boost);
    }
  }


  this.TachoBase = TachoBase;
  this.GaugeBuilder = GaugeBuilder;
  this.VelocityTachoApp = VelocityTachoApp;
  this.VelocityBoostApp = VelocityBoostApp;

  // The name of the font to use
  this.FONT_NAME = "Arial";
  // The angle range of the gauges
  this.GAUGE_RANGE = new TachoBase.Range(0.9*Math.PI, 2.1*Math.PI);


  return this;
}]);